#include "algorithm.hpp"
#include <iostream>
#include <fstream>
#include <sstream>
#include <stack>
#include <queue>

Graph ReadInGraph(std::string fileLocation) {
    std::ifstream graphFile(fileLocation);
    if (!graphFile.is_open()) {
        std::cerr << "Error opening file, check filename" << std::endl;
        exit(1);
    }
    Graph newGraph;
    graphFile >> newGraph.nodeCount;
    newGraph.vertices.resize(newGraph.nodeCount);
    std::string line;
    int source, destination, weight;
    std::getline(graphFile, line);
    while (std::getline(graphFile, line)) {
        std::istringstream iss(line);
        iss >> source;
        source--;
        do {
            iss >> destination >> weight;
            destination--;
            newGraph.vertices.at(source).nodeNumber = source;
            newGraph.AddEdge(source, destination, weight);
        } while (!iss.eof());
    }
    graphFile.close();
    return newGraph;
}

Graph::Graph(void) {
    nodeCount = 0;
}

Edge::Edge(Vertex* destination, int weight) {
    this->destination = destination;
    this->weight = weight;
} 

void Graph::AddEdge(int source, int destination, int weight) {
    vertices.at(source).edges.push_back(Edge(&(vertices.at(destination)), weight));
}

void Graph::PrintGraph(void) {
    for (int i = 0; i < vertices.size(); i++) {
        std::cout << "Source " << vertices.at(0).nodeNumber + 1 << " has ";
        for (Edge j : vertices.at(i).edges) { std::cout << " " << j.destination->nodeNumber + 1; }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}

std::vector<Vertex*> CleanSolution(std::vector<Vertex*> solution) {
    std::vector<Vertex*> solutionSanitized;
    int i = solution.size() - 2;
    Vertex* destination = solution.back();
    Vertex* source = destination;
    while (destination != solution.front()) {
        solutionSanitized.push_back(destination);
        while (source == destination && i > 0) {
            for (Edge e : solution.at(i)->edges) {
                if (e.destination->nodeNumber == destination->nodeNumber) {
                    source = solution.at(i);
                }
            }
            i--;
        }
    }
    solutionSanitized.push_back(destination);
    return solutionSanitized;
}

void PrintSolution(std::vector<Vertex*> solution) {
    int cost = 0;
    std::cout << "Path: " << solution.at(solution.size() - 1)->nodeNumber;
    for (int i = solution.size() - 2; i > 0; --i) {
        std::cout << "<-" << solution.at(i)->nodeNumber;
        cost += solution.at(i + 1)->edges.at(solution.at(i)->nodeNumber).weight;
    }
    std::cout << std::endl;
    std::cout << "Cost: " << cost << std::endl;
}

void UCBFS(Graph graph, int start) {
    std::queue<int> setds;
    std::vector<uint> dist(graph.nodeCount + 1, -1);
    setds.push(start);
    dist.at(start) = 0;
    while (!setds.empty())
    {
        int tmp = setds.front();
        setds.pop();
        for (Edge i : graph.vertices.at(tmp).edges)
        {
            int v = i.destination->nodeNumber;
            int weight = i.weight;
            if (dist[v] > dist.at(tmp) + weight)
            {
                if (dist[v] != ((uint)-1))
                    setds.pop();
                dist[v] = dist[tmp] + weight;
                setds.push(v);
            }
        }
    }
    std::cout << "Vertex\t Distance from Source" << std::endl;
    for (int i = 1; i < graph.nodeCount + 1; ++i) {
        std::cout << i << "\t\t" << dist[i] << std::endl;
    }
}

std::vector<Vertex*> BFS(Graph graph, int start, int end) {
    // Setup
    std::vector<Vertex*> solution;
    std::vector<bool> explored(graph.nodeCount + 1, 0);
    explored.at(start) = 1;
    std::queue<int> nodeQueue;
    nodeQueue.push(start);
    
    // Search
    while (!nodeQueue.empty()) {
        int currentNode = nodeQueue.front();
        nodeQueue.pop();
        solution.push_back(&(graph.vertices.at(currentNode)));
        if (currentNode == end) { break; }
        for (Edge i : graph.vertices.at(currentNode).edges) {
            if (!explored.at(i.destination->nodeNumber)) {
                explored.at(i.destination->nodeNumber) = 1;
                nodeQueue.push(i.destination->nodeNumber);
            }
        }
        if (solution.size() > graph.nodeCount + 1) { 
            std::cout << "Path too long" << std::endl;
            break;
        } 
    }
    return solution;
}

std::vector<Vertex*> DFS(Graph graph, int start, int end) {
    // Setup
    std::vector<Vertex*> solution;
    std::vector<bool> visited(graph.nodeCount, false);
    std::stack<Vertex*> vertexStack;
    Vertex* vertexCurrent;
    solution.emplace_back(&graph.vertices.at(start));
    vertexStack.push(&graph.vertices.at(start));

    // Search
    while (!vertexStack.empty()) {
        vertexCurrent = vertexStack.top();
        solution.push_back(vertexCurrent);
        vertexStack.pop();
        if (vertexCurrent->nodeNumber == end) { return solution; }
        for (Edge e : vertexCurrent->edges) {
            if (visited[e.destination->nodeNumber]) { continue; } // If visited
            visited[e.destination->nodeNumber] = true;
            for (auto v = e.destination->edges.rbegin(); v != e.destination->edges.rend(); ++v) {
                if (!visited[v->destination->nodeNumber]) { 
                    vertexStack.push(&graph.vertices.at(v->destination->nodeNumber)); 
                }
            }
        }
    }

    return solution;
}
